# SPDX-License-Identifier: ISC
# Copyright (c) Justus Winter <4winter@informatik.uni-hamburg.de>

from afew.filters.SentMailsFilter import SentMailsFilter
from afew.NotmuchSettings import get_notmuch_new_tags


class ArchiveSentMailsFilter(SentMailsFilter):
    message = 'Archiving all mails sent by myself to others'

    def __init__(self, database, sent_tag='', to_transforms=''):
        super().__init__(database, sent_tag)

    def handle_message(self, message):
        super().handle_message(message)
        self.remove_tags(message, *get_notmuch_new_tags())



# SPDX-License-Identifier: ISC
# Copyright (c) Justus Winter <4winter@informatik.uni-hamburg.de>

import configparser


class GetListMixIn:
    def get_list(self, section, key, delimiter=';',
                 filter_=lambda value: value.strip(),
                 include_falsish=False):
        result = (filter_(value)
                  for value in self.get(section, key).split(delimiter))

        if include_falsish:
            return result
        else:
            return filter(None, result)


class ConfigParser(configparser.ConfigParser, GetListMixIn): pass


class RawConfigParser(configparser.RawConfigParser, GetListMixIn): pass



# SPDX-License-Identifier: ISC
# Copyright (c) Justus Winter <4winter@informatik.uni-hamburg.de>

import re
from datetime import datetime


def get_message_summary(message):
    when = datetime.fromtimestamp(float(message.get_date()))
    sender = get_sender(message)
    subject = message.get_header('Subject')
    return '[{date}] {sender} | {subject}'.format(date=when, sender=sender,
                                                  subject=subject)


def get_sender(message):
    sender = message.get_header('From')
    name_match = re.search(r'(.+) <.+@.+\..+>', sender)
    if name_match:
        sender = name_match.group(1)
    return sender



# SPDX-License-Identifier: ISC
# Copyright (c) Justus Winter <4winter@informatik.uni-hamburg.de>

from afew.filters.HeaderMatchingFilter import HeaderMatchingFilter


class SpamFilter(HeaderMatchingFilter):
    message = 'Tagging spam messages'
    header = 'X-Spam-Flag'
    pattern = 'YES'

    def __init__(self, database, tags='+spam', spam_tag=None, **kwargs):
        if spam_tag is not None:
            # this is for backward-compatibility
            tags = '+' + spam_tag
        kwargs['tags'] = [tags]
        super().__init__(database, **kwargs)



# SPDX-License-Identifier: ISC
# Copyright (c) Justus Winter <4winter@informatik.uni-hamburg.de>

import pkg_resources

RAISEIT = object()


class FilterRegistry:
    """
    The FilterRegistry is responsible for returning
    filters by key.
    Filters get registered via entry points.
    To avoid any circular dependencies, the registry loads
    the Filters lazily
    """

    def __init__(self, filters):
        self._filteriterator = filters

    @property
    def filter(self):
        if not hasattr(self, '_filter'):
            self._filter = {}
            for f in self._filteriterator:
                self._filter[f.name] = f.load()
        return self._filter

    def get(self, key, default=RAISEIT):
        if default == RAISEIT:
            return self.filter[key]
        else:
            return self.filter.get(key, default)

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        self.filter[key] = value

    def __delitem__(self, key):
        del self.filter[key]

    def keys(self):
        return self.filter.keys()

    def values(self):
        return self.filter.values()

    def items(self):
        return self.filter.items()


all_filters = FilterRegistry(pkg_resources.iter_entry_points('afew.filter'))


def register_filter(klass):
    '''Decorator function for registering a class as a filter.'''

    all_filters[klass.__name__] = klass
    return klass



# SPDX-License-Identifier: ISC
# Copyright (c) Justus Winter <4winter@informatik.uni-hamburg.de>

import collections
import logging


class Filter:
    message = 'No message specified for filter'
    tags = []
    tags_blacklist = []

    def __init__(self, database, **kwargs):
        super().__init__()

        self.log = logging.getLogger('{}.{}'.format(
            self.__module__, self.__class__.__name__))

        self.database = database
        if 'tags' not in kwargs:
            kwargs['tags'] = self.tags
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.flush_changes()
        self._tags_to_add = []
        self._tags_to_remove = []
        for tag_action in self.tags:
            if tag_action[0] not in '+-':
                raise ValueError('Each tag must be preceded by either + or -')

            (self._tags_to_add if tag_action[0] == '+' else self._tags_to_remove).append(tag_action[1:])

        self._tag_blacklist = set(self.tags_blacklist)

    def flush_changes(self):
        '''
        (Re)Initializes the data structures that hold the enqueued
        changes to the notmuch database.
        '''
        self._add_tags = collections.defaultdict(lambda: set())
        self._remove_tags = collections.defaultdict(lambda: set())
        self._flush_tags = []

    def run(self, query):
        self.log.info(self.message)

        if getattr(self, 'query', None):
            if query:
                query = '(%s) AND (%s)' % (query, self.query)
            else:
                query = self.query

        for message in self.database.get_messages(query):
            self.handle_message(message)

    def handle_message(self, message):
        if not self._tag_blacklist.intersection(message.get_tags()):
            self.remove_tags(message, *self._tags_to_remove)
            self.add_tags(message, *self._tags_to_add)

    def add_tags(self, message, *tags):
        if tags:
            self.log.debug('Adding tags %s to id:%s' % (', '.join(tags),
                                                        message.get_message_id()))
            self._add_tags[message.get_message_id()].update(tags)

    def remove_tags(self, message, *tags):
        if tags:
            filtered_tags = list(tags)
            self.log.debug('Removing tags %s from id:%s' % (', '.join(filtered_tags),
                                                            message.get_message_id()))
            self._remove_tags[message.get_message_id()].update(filtered_tags)

    def flush_tags(self, message):
        self.log.debug('Removing all tags from id:%s' %
                       message.get_message_id())
        self._flush_tags.append(message.get_message_id())

    def commit(self, dry_run=True):
        dirty_messages = set()
        dirty_messages.update(self._flush_tags)
        dirty_messages.update(self._add_tags.keys())
        dirty_messages.update(self._remove_tags.keys())

        if not dirty_messages:
            return

        if dry_run:
            self.log.info('I would commit changes to %i messages' % len(dirty_messages))
        else:
            self.log.info('Committing changes to %i messages' % len(dirty_messages))
            db = self.database.open(rw=True)

            for message_id in dirty_messages:
                message = db.find_message(message_id)

                if message_id in self._flush_tags:
                    message.remove_all_tags()

                for tag in self._add_tags.get(message_id, []):
                    message.add_tag(tag)

                for tag in self._remove_tags.get(message_id, []):
                    message.remove_tag(tag)

        self.flush_changes()



# SPDX-License-Identifier: ISC
# Copyright (c) dtk <dtk@gmx.de>

from afew.filters.BaseFilter import Filter
from afew.NotmuchSettings import notmuch_settings
import re
import shlex


class FolderNameFilter(Filter):
    message = 'Tags all new messages with their folder'

    def __init__(self, database, folder_blacklist='', folder_transforms='',
                 maildir_separator='.', folder_explicit_list='', folder_lowercases=''):
        super().__init__(database)

        self.__filename_pattern = '{mail_root}/(?P<maildirs>.*)/(cur|new)/[^/]+'.format(
            mail_root=notmuch_settings.get('database', 'path').rstrip('/'))
        self.__folder_explicit_list = set(shlex.split(folder_explicit_list))
        self.__folder_blacklist = set(shlex.split(folder_blacklist))
        self.__folder_transforms = self.__parse_transforms(folder_transforms)
        self.__folder_lowercases = folder_lowercases != ''
        self.__maildir_separator = maildir_separator

    def handle_message(self, message):
        # Find all the dirs in the mail directory that this message
        # belongs to
        maildirs = [re.match(self.__filename_pattern, filename)
                    for filename in message.get_filenames()]
        maildirs = filter(None, maildirs)
        if maildirs:
            # Make the folders relative to mail_root and split them.
            folder_groups = [maildir.group('maildirs').split(self.__maildir_separator)
                             for maildir in maildirs]
            folders = set([folder
                           for folder_group in folder_groups
                           for folder in folder_group])
            self.log.debug('found folders {} for message {!r}'.format(
                folders, message.get_header('subject')))

            # remove blacklisted folders
            clean_folders = folders - self.__folder_blacklist
            if self.__folder_explicit_list:
                # only explicitly listed folders
                clean_folders &= self.__folder_explicit_list
            # apply transformations
            transformed_folders = self.__transform_folders(clean_folders)

            self.add_tags(message, *transformed_folders)

    def __transform_folders(self, folders):
        """
        Transforms the given collection of folders according to the transformation rules.
        """
        transformations = set()
        for folder in folders:
            if folder in self.__folder_transforms:
                transformations.add(self.__folder_transforms[folder])
            else:
                transformations.add(folder)
        if self.__folder_lowercases:
            rtn = set()
            for folder in transformations:
                rtn.add(folder.lower())
            return rtn
        return transformations

    def __parse_transforms(self, transformation_description):
        """
        Parses the transformation rules specified in the config file.
        """
        transformations = dict()
        for rule in shlex.split(transformation_description):
            folder, tag = rule.split(':')
            transformations[folder] = tag
        return transformations



# SPDX-License-Identifier: ISC
# Copyright (c) Amadeusz Zolnowski <aidecoe@aidecoe.name>

"""
DKIM validator filter.

Verifies DKIM signature of an e-mail which has DKIM header.
"""

import logging

import dkim
import dns.exception

from afew.filters.BaseFilter import Filter


class DKIMVerifyError(Exception):
    """Failed to verify DKIM signature.
    """


def verify_dkim(path):
    """
    Verify DKIM signature of an e-mail file.

    :param path: Path to the e-mail file.
    :returns: Whether DKIM signature is valid or not.
    """
    with open(path, 'rb') as message_file:
        message_bytes = message_file.read()

    try:
        return dkim.verify(message_bytes)
    except (dns.exception.DNSException, dkim.DKIMException) as exception:
        raise DKIMVerifyError(str(exception)) from exception


class DKIMValidityFilter(Filter):
    """
    Verifies DKIM signature of an e-mail which has DKIM header.
    """
    message = 'Verify DKIM signature'
    header = 'DKIM-Signature'

    def __init__(self, database, ok_tag='dkim-ok', fail_tag='dkim-fail'):
        super().__init__(database)
        self.dkim_tag = {True: ok_tag, False: fail_tag}
        self.log = logging.getLogger('{}.{}'.format(
            self.__module__, self.__class__.__name__))

    def handle_message(self, message):
        if message.get_header(self.header):
            try:
                dkim_ok = all(map(verify_dkim, message.get_filenames()))
            except DKIMVerifyError as verify_error:
                self.log.warning(
                    "Failed to verify DKIM of '%s': %s "
                    "(marked as 'dkim-fail')",
                    message.get_message_id(),
                    verify_error
                )
                dkim_ok = False
            self.add_tags(message, self.dkim_tag[dkim_ok])



# SPDX-License-Identifier: ISC
# Copyright (c) 2012 Justus Winter <4winter@informatik.uni-hamburg.de>
# Copyright (c) 2013 Patrick Gerken <do3cc@patrick-gerken.de>
# Copyright (c) 2013 Patrick Totzke <patricktotzke@gmail.com>
# Copyright (c) 2014 Lars Kellogg-Stedman <lars@redhat.com>

from afew.filters.BaseFilter import Filter

import re


class HeaderMatchingFilter(Filter):
    message = 'Tagging based on specific header values matching a given RE'
    header = None
    pattern = None

    def __init__(self, database, **kwargs):
        super().__init__(database, **kwargs)
        if self.pattern is not None:
            self.pattern = re.compile(self.pattern, re.I)

    def handle_message(self, message):
        if self.header is not None and self.pattern is not None:
            if not self._tag_blacklist.intersection(message.get_tags()):
                value = message.get_header(self.header)
                match = self.pattern.search(value)
                if match:
                    sub = (lambda tag:
                           tag.format(**match.groupdict()).lower())
                    self.remove_tags(message, *map(sub, self._tags_to_remove))
                    self.add_tags(message, *map(sub, self._tags_to_add))



# SPDX-License-Identifier: ISC
# Copyright (c) Justus Winter <4winter@informatik.uni-hamburg.de>

from afew.filters.HeaderMatchingFilter import HeaderMatchingFilter


class ListMailsFilter(HeaderMatchingFilter):
    message = 'Tagging mailing list posts'
    query = 'NOT tag:lists'
    pattern = r"<(?P<list_id>[a-z0-9!#$%&'*+/=?^_`{|}~-]+)\."
    header = 'List-Id'
    tags = ['+lists', '+lists/{list_id}']



# SPDX-License-Identifier: ISC
# Copyright (c) Justus Winter <4winter@informatik.uni-hamburg.de>

import glob
import sys
import logging
import argparse

from afew.Database import Database
from afew.main import main as inner_main
from afew.FilterRegistry import all_filters
from afew.Settings import user_config_dir, get_filter_chain, \
    get_mail_move_rules, get_mail_move_age, get_mail_move_rename
from afew.NotmuchSettings import read_notmuch_settings, get_notmuch_new_query
from afew.version import version

parser = argparse.ArgumentParser()
parser.add_argument('-V', '--version', action='version', version=version)

# the actions
action_group = parser.add_argument_group(
    'Actions',
    'Please specify exactly one action.'
)
action_group.add_argument(
    '-t', '--tag', action='store_true',
    help='run the tag filters'
)
action_group.add_argument(
    '-w', '--watch', action='store_true',
    help='continuously monitor the mailbox for new files'
)
action_group.add_argument(
    '-m', '--move-mails', action='store_true',
    help='move mail files between maildir folders'
)

# query modifiers
query_modifier_group = parser.add_argument_group(
    'Query modifiers',
    'Please specify either --all or --new or a query string.'
)
query_modifier_group.add_argument(
    '-a', '--all', action='store_true',
    help='operate on all messages'
)
query_modifier_group.add_argument(
    '-n', '--new', action='store_true',
    help='operate on all new messages'
)
query_modifier_group.add_argument(
    'query', nargs='*', help='a notmuch query to find messages to work on'
)

# general options
options_group = parser.add_argument_group('General options')
# TODO: get config via notmuch api
options_group.add_argument(
    '-C', '--notmuch-config', default=None,
    help='path to the notmuch configuration file [default: $NOTMUCH_CONFIG or'
         ' ~/.notmuch-config]'
)
options_group.add_argument(
    '-e', '--enable-filters',
    help="filter classes to use, separated by ',' [default: filters specified"
         " in afew's config]"
)
options_group.add_argument(
    '-d', '--dry-run', default=False, action='store_true',
    help="don't change the db [default: %(default)s]"
)
options_group.add_argument(
    '-R', '--reference-set-size', type=int, default=1000,
    help='size of the reference set [default: %(default)s]'
)

options_group.add_argument(
    '-T', '--reference-set-timeframe', type=int, default=30, metavar='DAYS',
    help='do not use mails older than DAYS days [default: %(default)s]'
)

options_group.add_argument(
    '-v', '--verbose', dest='verbosity', action='count', default=0,
    help='be more verbose, can be given multiple times'
)

options_group.add_argument(
    '-N', '--notmuch-args', default='',
    help='arguments for notmuch new (in move mode)'
)


def main():
    if sys.version_info < (3, 6):
        sys.exit("Python 3.6 or later is required.")

    args = parser.parse_args()

    no_actions = len(list(filter(None, (
        args.tag,
        args.watch,
        args.move_mails
    ))))
    if no_actions == 0:
        sys.exit('You need to specify an action')
    elif no_actions > 1:
        sys.exit('Please specify exactly one action')

    no_query_modifiers = len(list(filter(None, (args.all,
                                                args.new, args.query))))
    if no_query_modifiers == 0 and not args.watch \
            and not args.move_mails:
        sys.exit('You need to specify one of --new, --all or a query string')
    elif no_query_modifiers > 1:
        sys.exit('Please specify either --all, --new or a query string')

    read_notmuch_settings(args.notmuch_config)

    if args.new:
        query_string = get_notmuch_new_query()
    elif args.all:
        query_string = ''
    else:
        query_string = ' '.join(args.query)

    loglevel = {
        0: logging.WARNING,
        1: logging.INFO,
        2: logging.DEBUG,
    }[min(2, args.verbosity)]
    logging.basicConfig(level=loglevel)

    sys.path.insert(0, user_config_dir)
    for file_name in glob.glob1(user_config_dir, '*.py'):
        logging.info('Importing user filter %r' % (file_name,))
        __import__(file_name[:-3], level=0)

    if args.move_mails:
        args.mail_move_rules = get_mail_move_rules()
        args.mail_move_age = get_mail_move_age()
        args.mail_move_rename = get_mail_move_rename()

    with Database() as database:
        configured_filter_chain = get_filter_chain(database)
        if args.enable_filters:
            args.enable_filters = args.enable_filters.split(',')

            all_filters_set = set(all_filters.keys())
            enabled_filters_set = set(args.enable_filters)
            if not all_filters_set.issuperset(enabled_filters_set):
                sys.exit('Unknown filter(s) selected: %s' % (' '.join(
                    enabled_filters_set.difference(all_filters_set))))

            args.enable_filters = [all_filters[filter_name](database)
                                   for filter_name
                                   in args.enable_filters]
        else:
            args.enable_filters = configured_filter_chain

        inner_main(args, database, query_string)



# SPDX-License-Identifier: ISC

import email.message
from email.utils import make_msgid
from freezegun import freeze_time
import mailbox
import notmuch
import os
import shutil
import tempfile
import unittest

from afew.Database import Database
from afew.NotmuchSettings import notmuch_settings, write_notmuch_settings

def create_mail(msg, maildir, notmuch_db, tags, old=False):
    email_message = email.message.EmailMessage()
    # freezegun doesn't handle time zones properly when generating UNIX
    # timestamps.  When the local timezone is UTC+2, the generated timestamp
    # is 2 hours ahead of what it should be.  Due to this we need to make sure
    # that the dates are always sufficiently far behind 2019-01-30 12:00 to
    # handle up to UTC+12 .
    if old:
        email_message['Date'] = 'Wed, 10 Jan 2019 13:00:00 +0100'
    else:
        email_message['Date'] = 'Wed, 20 Jan 2019 13:00:00 +0100'
    email_message['From'] = 'You <you@example.org>'
    email_message['To'] = 'Me <me@example.com>'
    email_message['Message-ID'] = make_msgid()
    email_message.set_content(msg)

    maildir_message = mailbox.MaildirMessage(email_message)
    message_key = maildir.add(maildir_message)

    fname = os.path.join(maildir._path, maildir._lookup(message_key))
    notmuch_msg = notmuch_db.add_message(fname)
    for tag in tags:
        notmuch_msg.add_tag(tag, False)

    # Remove the angle brackets automatically added around the message ID by make_msgid.
    stripped_msgid = email_message['Message-ID'].strip('<>')
    return (stripped_msgid, msg)


@freeze_time("2019-01-30 12:00:00")
class TestMailMover(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

        os.environ['MAILDIR'] = self.test_dir
        os.environ['NOTMUCH_CONFIG'] = os.path.join(self.test_dir, 'notmuch-config')

        notmuch_settings['database'] = {'path': self.test_dir}
        notmuch_settings['new'] = {'tags': 'new'}
        write_notmuch_settings()

        # Create notmuch database
        Database().open(create=True).close()

        self.root = mailbox.Maildir(self.test_dir)
        self.inbox = self.root.add_folder('inbox')
        self.archive = self.root.add_folder('archive')
        self.spam = self.root.add_folder('spam')

        # Dict of rules that are passed to MailMover.
        #
        # The top level key represents a particular mail directory to work on.
        #
        # The second level key is the notmuch query that MailMover will execute,
        # and its value is the directory to move the matching emails to.
        self.rules = {
            '.inbox': {
                'tag:archive AND NOT tag:spam': '.archive',
                'tag:spam': '.spam',
            },
            '.archive': {
                'NOT tag:archive AND NOT tag:spam': '.inbox',
                'tag:spam': '.spam',
            },
            '.spam': {
                'NOT tag:spam AND tag:archive': '.archive',
                'NOT tag:spam AND NOT tag:archive': '.inbox',
            },
        }


    def tearDown(self):
        shutil.rmtree(self.test_dir)


    @staticmethod
    def get_folder_content(db, folder):
        return {
            (os.path.basename(msg.get_message_id()), msg.get_part(1).decode())
            for msg in db.do_query('folder:{}'.format(folder)).search_messages()
        }


    def test_all_rule_cases(self):
        from afew import MailMover

        with Database() as db:
            expect_inbox = set([
                create_mail('In inbox, untagged\n', self.inbox, db, []),
                create_mail('In archive, untagged\n', self.archive, db, []),
                create_mail('In spam, untagged\n', self.spam, db, []),
            ])

            expect_archive = set([
                create_mail('In inbox, tagged archive\n', self.inbox, db, ['archive']),
                create_mail('In archive, tagged archive\n', self.archive, db, ['archive']),
                create_mail('In spam, tagged archive\n', self.spam, db, ['archive']),
            ])

            expect_spam = set([
                create_mail('In inbox, tagged spam\n', self.inbox, db, ['spam']),
                create_mail('In inbox, tagged archive, spam\n', self.inbox, db, ['archive', 'spam']),
                create_mail('In archive, tagged spam\n', self.archive, db, ['spam']),
                create_mail('In archive, tagged archive, spam\n', self.archive, db, ['archive', 'spam']),
                create_mail('In spam, tagged spam\n', self.spam, db, ['spam']),
                create_mail('In spam, tagged archive, spam\n', self.spam, db, ['archive', 'spam']),
            ])

        mover = MailMover.MailMover(quiet=True)
        mover.move('.inbox', self.rules['.inbox'])
        mover.move('.archive', self.rules['.archive'])
        mover.move('.spam', self.rules['.spam'])
        mover.close()

        with Database() as db:
            self.assertEqual(expect_inbox, self.get_folder_content(db, '.inbox'))
            self.assertEqual(expect_archive, self.get_folder_content(db, '.archive'))
            self.assertEqual(expect_spam, self.get_folder_content(db, '.spam'))


    def test_max_age(self):
        from afew import MailMover

        with Database() as db:
            expect_inbox = set([
                create_mail('In inbox, tagged archive, old\n', self.inbox, db, ['archive'], old=True),
            ])

            expect_archive = set([
                create_mail('In inbox, tagged archive\n', self.inbox, db, ['archive']),
            ])

            expect_spam = set([])

        mover = MailMover.MailMover(max_age=15, quiet=True)
        mover.move('.inbox', self.rules['.inbox'])
        mover.move('.archive', self.rules['.archive'])
        mover.move('.spam', self.rules['.spam'])
        mover.close()

        with Database() as db:
            self.assertEqual(expect_inbox, self.get_folder_content(db, '.inbox'))
            self.assertEqual(expect_archive, self.get_folder_content(db, '.archive'))
            self.assertEqual(expect_spam, self.get_folder_content(db, '.spam'))



# SPDX-License-Identifier: ISC
# Copyright (c) Amadeusz Zolnowski <aidecoe@aidecoe.name>

import re

from afew.filters.BaseFilter import Filter
from afew.NotmuchSettings import notmuch_settings


class MeFilter(Filter):
    message = 'Tagging all mails sent directly to myself'
    _bare_email_re = re.compile(r"[^<]*<(?P<email>[^@<>]+@[^@<>]+)>")

    def __init__(self, database, me_tag='to-me', tags_blacklist=[]):
        super().__init__(database, tags_blacklist=tags_blacklist)

        my_addresses = set()
        my_addresses.add(notmuch_settings.get('user', 'primary_email'))
        if notmuch_settings.has_option('user', 'other_email'):
            for other_email in notmuch_settings.get_list('user', 'other_email'):
                my_addresses.add(other_email)

        self.query = ' OR '.join('to:"%s"' % address
                                 for address in my_addresses)

        self.me_tag = me_tag

    def handle_message(self, message):
        if not self._tag_blacklist.intersection(message.get_tags()):
            self.add_tags(message, self.me_tag)



# Copyright (c) Jens Neuhalfen <jens@neuhalfen.name>

from afew.filters.BaseFilter import Filter


class PropagateTagsInThreadFilter(Filter):
    """
    Propagate tags in threads. For each new message the mail thread is examined. If one of the
    configured tags is found, it is automatically attached to the new message.

    This enables a very easy workflow where entire threads can be tagged automatically.

    Config:

    [PropagateTagsInThreadFilter]
    propagate_tags = "project_A;billing;private"
    filter = not is:spam

    """

    def handle_message(self, message):
        for tag in self._propagate_tags:
            tag_query = 'thread:"%s" AND is:"%s"' % (message.get_thread_id(), tag)
            if self._filter:
                query = self.database.get_messages("(%s) AND (%s)" % (tag_query, self._filter))
            else:
                query = self.database.get_messages(tag_query)

            if len(list(query)):
                self.add_tags(message, tag)

    def __init__(self, database, propagate_tags="", filter=None, **kwargs):
        if filter:
            self.message = "Propagating tag(s) '%s' for messages matching '%s' to whole threads" % (
                propagate_tags, filter)
        else:
            self.message = "Propagating tag(s) '%s' to whole threads" % (propagate_tags,)

        self._filter = filter
        self._propagate_tags = [t for t in propagate_tags.split(";") if len(t) > 0]

        super(PropagateTagsInThreadFilter, self).__init__(database, **kwargs)



# SPDX-License-Identifier: ISC
# Copyright (c) Justus Winter <4winter@informatik.uni-hamburg.de>

from afew.filters.BaseFilter import Filter
from afew.NotmuchSettings import get_notmuch_new_tags, get_notmuch_new_query


class InboxFilter(Filter):
    message = 'Retags all messages not tagged as junk or killed as inbox'
    tags = ['+inbox']
    tags_blacklist = ['killed', 'spam']

    @property
    def query(self):
        '''
        Need to read the notmuch settings first. Using a property here
        so that the setting is looked up on demand.
        '''
        return get_notmuch_new_query()

    def handle_message(self, message):
        self.remove_tags(message, *get_notmuch_new_tags())
        super().handle_message(message)



# SPDX-License-Identifier: ISC
# Copyright (c) dtk <dtk@gmx.de>

import logging
import os
import shutil
import uuid
from datetime import date, datetime, timedelta
from subprocess import check_call, CalledProcessError, DEVNULL

import notmuch

from afew.Database import Database
from afew.utils import get_message_summary


class MailMover(Database):
    """
    Move mail files matching a given notmuch query into a target maildir folder.
    """

    def __init__(self, max_age=0, rename=False, dry_run=False, notmuch_args='', quiet=False):
        super().__init__()
        self.db = notmuch.Database(self.db_path)
        self.query = 'folder:"{folder}" AND {subquery}'
        if max_age:
            days = timedelta(int(max_age))
            start = date.today() - days
            now = datetime.now()
            self.query += ' AND {start}..{now}'.format(start=start.strftime('%s'),
                                                       now=now.strftime('%s'))
        self.dry_run = dry_run
        self.rename = rename
        self.notmuch_args = notmuch_args
        self.quiet = quiet

    def get_new_name(self, fname, destination):
        basename = os.path.basename(fname)
        submaildir = os.path.split(os.path.split(fname)[0])[1]
        if self.rename:
            parts = basename.split(':')
            if len(parts) > 1:
                flagpart = ':' + parts[-1]
            else:
                flagpart = ''
                # construct a new filename, composed of a made-up ID and the flags part
                # of the original filename.
            basename = str(uuid.uuid1()) + flagpart
        return os.path.join(destination, submaildir, basename)

    def move(self, maildir, rules):
        """
        Move mails in folder maildir according to the given rules.
        """
        # identify and move messages
        logging.info("checking mails in '{}'".format(maildir))
        to_delete_fnames = []
        moved = False
        for query in rules.keys():
            destination = '{}/{}/'.format(self.db_path, rules[query])
            main_query = self.query.format(
                folder=maildir.replace("\"", "\\\""), subquery=query)
            logging.debug("query: {}".format(main_query))
            messages = notmuch.Query(self.db, main_query).search_messages()
            for message in messages:
                # a single message (identified by Message-ID) can be in several
                # places; only touch the one(s) that exists in this maildir
                all_message_fnames = message.get_filenames()
                to_move_fnames = [name for name in all_message_fnames
                                  if maildir in name]
                if not to_move_fnames:
                    continue
                moved = True
                self.__log_move_action(message, maildir, rules[query],
                                       self.dry_run)
                for fname in to_move_fnames:
                    if self.dry_run:
                        continue
                    try:
                        shutil.copy2(fname, self.get_new_name(fname, destination))
                        to_delete_fnames.append(fname)
                    except shutil.SameFileError:
                        logging.warn("trying to move '{}' onto itself".format(fname))
                        continue
                    except shutil.Error as e:
                        # this is ugly, but shutil does not provide more
                        # finely individuated errors
                        if str(e).endswith("already exists"):
                            continue
                        else:
                            raise

        # remove mail from source locations only after all copies are finished
        for fname in set(to_delete_fnames):
            os.remove(fname)

        # update notmuch database
        if not self.dry_run:
            if moved:
                logging.info("updating database")
                self.__update_db(maildir)
        else:
            logging.info("Would update database")

    def __update_db(self, maildir):
        """
        Update the database after mail files have been moved in the filesystem.
        """
        try:
            if self.quiet:
                check_call(['notmuch', 'new'] + self.notmuch_args.split(), stdout=DEVNULL, stderr=DEVNULL)
            else:
                check_call(['notmuch', 'new'] + self.notmuch_args.split())
        except CalledProcessError as err:
            logging.error("Could not update notmuch database "
                          "after syncing maildir '{}': {}".format(maildir, err))
            raise SystemExit

    def __log_move_action(self, message, source, destination, dry_run):
        '''
        Report which mails have been identified for moving.
        '''
        if not dry_run:
            level = logging.DEBUG
            prefix = 'moving mail'
        else:
            level = logging.INFO
            prefix = 'I would move mail'
        logging.log(level, prefix)
        logging.log(level, "    {}".format(get_message_summary(message).encode('utf8')))
        logging.log(level, "from '{}' to '{}'".format(source, destination))



# SPDX-License-Identifier: ISC
# Copyright (c) Justus Winter <4winter@informatik.uni-hamburg.de>

import os

from afew.configparser import RawConfigParser

notmuch_settings = RawConfigParser()


def read_notmuch_settings(path=None):
    if path == None:
        path = os.environ.get('NOTMUCH_CONFIG', os.path.expanduser('~/.notmuch-config'))

    with open(path) as fp:
        notmuch_settings.read_file(fp)

def write_notmuch_settings(path = None):
    if path == None:
        path = os.environ.get('NOTMUCH_CONFIG', os.path.expanduser('~/.notmuch-config'))

    with open(path, 'w+') as fp:
        notmuch_settings.write(fp)


def get_notmuch_new_tags():
    # see issue 158
    return filter(lambda x: x != 'unread', notmuch_settings.get_list('new', 'tags'))


def get_notmuch_new_query():
    return '(%s)' % ' AND '.join('tag:%s' % tag for tag in get_notmuch_new_tags())



"""Test suite for DKIMValidityFilter.
"""
import unittest
from email.utils import make_msgid
from unittest import mock

import dkim
import dns.exception

from afew.Database import Database
from afew.filters.DKIMValidityFilter import DKIMValidityFilter


class _AddTags:  # pylint: disable=too-few-public-methods
    """Mock for `add_tags` method of base filter. We need to easily collect
    tags added by filter for test assertion.
    """
    def __init__(self, tags):
        self._tags = tags

    def __call__(self, message, *tags):
        self._tags.update(tags)


def _make_dkim_validity_filter():
    """Make `DKIMValidityFilter` with mocked `DKIMValidityFilter.add_tags`
    method, so in tests we can easily check what tags were added by filter
    without fiddling with db.
    """
    tags = set()
    add_tags = _AddTags(tags)
    dkim_filter = DKIMValidityFilter(Database())
    dkim_filter.add_tags = add_tags
    return dkim_filter, tags


def _make_message():
    """Make mock email Message.

    Mocked methods:

    - `get_header()` returns non-empty string. When testing with mocked
      function for verifying DKIM signature, DKIM signature doesn't matter as
      long as it's non-empty string.

    - `get_filenames()` returns list of non-empty string. When testing with
    mocked file open, it must just be non-empty string.

    - `get_message_id()` returns some generated message ID.
    """
    message = mock.Mock()
    message.get_header.return_value = 'sig'
    message.get_filenames.return_value = ['a']
    message.get_message_id.return_value = make_msgid()
    return message


class TestDKIMValidityFilter(unittest.TestCase):
    """Test suite for `DKIMValidityFilter`.
    """
    @mock.patch('afew.filters.DKIMValidityFilter.open',
                mock.mock_open(read_data=b''))
    def test_no_dkim_header(self):
        """Test message without DKIM-Signature header doesn't get any tags.
        """
        dkim_filter, tags = _make_dkim_validity_filter()
        message = _make_message()
        message.get_header.return_value = False

        with mock.patch('afew.filters.DKIMValidityFilter.dkim.verify') \
                as dkim_verify:
            dkim_verify.return_value = True
            dkim_filter.handle_message(message)

        self.assertSetEqual(tags, set())

    @mock.patch('afew.filters.DKIMValidityFilter.open',
                mock.mock_open(read_data=b''))
    def test_dkim_all_ok(self):
        """Test message, with multiple files all having good signature, gets
        only 'dkim-ok' tag.
        """
        dkim_filter, tags = _make_dkim_validity_filter()
        message = _make_message()
        message.get_filenames.return_value = ['a', 'b', 'c']

        with mock.patch('afew.filters.DKIMValidityFilter.dkim.verify') \
                as dkim_verify:
            dkim_verify.return_value = True
            dkim_filter.handle_message(message)

        self.assertSetEqual(tags, {'dkim-ok'})

    @mock.patch('afew.filters.DKIMValidityFilter.open',
                mock.mock_open(read_data=b''))
    def test_dkim_all_fail(self):
        """Test message, with multiple files all having bad signature, gets
        only 'dkim-fail' tag.
        """
        dkim_filter, tags = _make_dkim_validity_filter()
        message = _make_message()
        message.get_filenames.return_value = ['a', 'b', 'c']

        with mock.patch('afew.filters.DKIMValidityFilter.dkim.verify') \
                as dkim_verify:
            dkim_verify.return_value = False
            dkim_filter.handle_message(message)

        self.assertSetEqual(tags, {'dkim-fail'})

    @mock.patch('afew.filters.DKIMValidityFilter.open',
                mock.mock_open(read_data=b''))
    def test_dkim_some_fail(self):
        """Test message, with multiple files but only some having bad
        signature, still gets only 'dkim-fail' tag.
        """
        dkim_filter, tags = _make_dkim_validity_filter()
        message = _make_message()
        message.get_filenames.return_value = ['a', 'b', 'c']

        with mock.patch('afew.filters.DKIMValidityFilter.dkim.verify') \
                as dkim_verify:
            dkim_verify.side_effect = [True, False, True]
            dkim_filter.handle_message(message)

        self.assertSetEqual(tags, {'dkim-fail'})

    @mock.patch('afew.filters.DKIMValidityFilter.open',
                mock.mock_open(read_data=b''))
    def test_dkim_dns_resolve_failure(self):
        """Test message, on which DNS resolution failure happens when verifying
        DKIM signature, gets only 'dkim-fail' tag.
        """
        dkim_filter, tags = _make_dkim_validity_filter()
        message = _make_message()

        with mock.patch('afew.filters.DKIMValidityFilter.dkim.verify') \
                as dkim_verify:
            dkim_verify.side_effect = dns.resolver.NoNameservers()
            dkim_filter.handle_message(message)

        self.assertSetEqual(tags, {'dkim-fail'})

    @mock.patch('afew.filters.DKIMValidityFilter.open',
                mock.mock_open(read_data=b''))
    def test_dkim_verify_failed(self):
        """Test message, on which DKIM key parsing failure occurs, gets only
        'dkim-fail' tag.
        """
        dkim_filter, tags = _make_dkim_validity_filter()
        message = _make_message()

        with mock.patch('afew.filters.DKIMValidityFilter.dkim.verify') \
                as dkim_verify:
            dkim_verify.side_effect = dkim.KeyFormatError()
            dkim_filter.handle_message(message)

        self.assertSetEqual(tags, {'dkim-fail'})



# SPDX-License-Identifier: ISC
# Copyright (c) Justus Winter <4winter@informatik.uni-hamburg.de>

from afew.filters.BaseFilter import Filter


class KillThreadsFilter(Filter):
    message = 'Looking for messages in killed threads that are not yet killed'
    query = 'NOT tag:killed'

    def handle_message(self, message):
        query = self.database.get_messages('thread:"%s" AND tag:killed' % message.get_thread_id())

        if len(list(query)):
            self.add_tags(message, 'killed')



# SPDX-License-Identifier: ISC
# Copyright (c) Justus Winter <4winter@informatik.uni-hamburg.de>

import os
import re
import stat
import logging
import platform
import queue
import threading

if platform.system() != 'Linux':
    raise ImportError('Unsupported platform: {!r}'.format(platform.system()))

import notmuch
import pyinotify


class EventHandler(pyinotify.ProcessEvent):
    def __init__(self, options, database):
        self.options = options
        self.database = database
        super().__init__()

    ignore_re = re.compile('(/xapian/.*(base.|tmp)$)|(\.lock$)|(/dovecot)')

    def process_IN_DELETE(self, event):
        if self.ignore_re.search(event.pathname):
            return

        logging.debug("Detected file removal: {!r}".format(event.pathname))
        self.database.remove_message(event.pathname)
        self.database.close()

    def process_IN_MOVED_TO(self, event):
        if self.ignore_re.search(event.pathname):
            return

        src_pathname = event.src_pathname if hasattr(event, 'src_pathname') else None
        logging.debug("Detected file rename: {!r} -> {!r}".format(src_pathname, event.pathname))

        def new_mail(message):
            for filter_ in self.options.enable_filters:
                try:
                    filter_.run('id:"{}"'.format(message.get_message_id()))
                    filter_.commit(self.options.dry_run)
                except Exception as e:
                    logging.warn('Error processing mail with filter {!r}: {}'.format(filter_.message, e))

        try:
            self.database.add_message(event.pathname,
                                      sync_maildir_flags=True,
                                      new_mail_handler=new_mail)
        except notmuch.FileError as e:
            logging.warn('Error opening mail file: {}'.format(e))
            return
        except notmuch.FileNotEmailError as e:
            logging.warn('File does not look like an email: {}'.format(e))
            return
        else:
            if src_pathname:
                self.database.remove_message(src_pathname)
        finally:
            self.database.close()


def watch_for_new_files(options, database, paths, daemonize=False):
    wm = pyinotify.WatchManager()
    mask = (
            pyinotify.IN_DELETE |
            pyinotify.IN_MOVED_FROM |
            pyinotify.IN_MOVED_TO)
    handler = EventHandler(options, database)
    notifier = pyinotify.Notifier(wm, handler)

    logging.debug('Registering inotify watch descriptors')
    wdds = dict()
    for path in paths:
        wdds[path] = wm.add_watch(path, mask)

    # TODO: honor daemonize
    logging.debug('Running mainloop')
    notifier.loop()


import ctypes
import contextlib

try:
    libc = ctypes.CDLL(ctypes.util.find_library("c"))
except ImportError as e:
    raise ImportError('Could not load libc: {}'.format(e))


class Libc:
    class c_dir(ctypes.Structure):
        pass

    c_dir_p = ctypes.POINTER(c_dir)

    opendir = libc.opendir
    opendir.argtypes = [ctypes.c_char_p]
    opendir.restype = c_dir_p

    closedir = libc.closedir
    closedir.argtypes = [c_dir_p]
    closedir.restype = ctypes.c_int

    @classmethod
    @contextlib.contextmanager
    def open_directory(cls, path):
        handle = cls.opendir(path)
        yield handle
        cls.closedir(handle)

    class c_dirent(ctypes.Structure):
        '''
        man 3 readdir says::

        On Linux, the dirent structure is defined as follows:

           struct dirent {
               ino_t          d_ino;       /* inode number */
               off_t          d_off;       /* offset to the next dirent */
               unsigned short d_reclen;    /* length of this record */
               unsigned char  d_type;      /* type of file; not supported
                                              by all file system types */
               char           d_name[256]; /* filename */
           };
        '''
        _fields_ = (
            ('d_ino', ctypes.c_long),
            ('d_off', ctypes.c_long),
            ('d_reclen', ctypes.c_ushort),
            ('d_type', ctypes.c_byte),
            ('d_name', ctypes.c_char * 4096),
        )

    c_dirent_p = ctypes.POINTER(c_dirent)

    readdir = libc.readdir
    readdir.argtypes = [c_dir_p]
    readdir.restype = c_dirent_p

    # magic value for directory
    DT_DIR = 4


blacklist = {'.', '..', 'tmp'}


def walk_linux(channel, path):
    channel.put(path)

    with Libc.open_directory(path) as handle:
        while True:
            dirent_p = Libc.readdir(handle)
            if not dirent_p:
                break

            if dirent_p.contents.d_type == Libc.DT_DIR and \
                    dirent_p.contents.d_name not in blacklist:
                walk_linux(channel, os.path.join(path, dirent_p.contents.d_name))


def walk(channel, path):
    channel.put(path)

    for child_path in (os.path.join(path, child)
                       for child in os.listdir(path)
                       if child not in blacklist):
        try:
            stat_result = os.stat(child_path)
        except:
            continue

        if stat_result.st_mode & stat.S_IFDIR:
            walk(channel, child_path)


def walker(channel, path):
    walk_linux(channel, path)
    channel.put(None)


def quick_find_dirs_hack(path):
    results = queue.Queue()

    walker_thread = threading.Thread(target=walker, args=(results, path))
    walker_thread.daemon = True
    walker_thread.start()

    while True:
        result = results.get()

        if result != None:
            yield result
        else:
            break



# SPDX-License-Identifier: ISC
# Copyright (c) Justus Winter <4winter@informatik.uni-hamburg.de>

import os
import time
import logging

import notmuch

from afew.NotmuchSettings import notmuch_settings, get_notmuch_new_tags


class Database:
    """
    Convenience wrapper around `notmuch`.
    """

    def __init__(self):
        self.db_path = self._calculate_db_path()
        self.handle = None

    def _calculate_db_path(self):
        """
        Calculates the path to use for the database. Supports notmuch's
        methodology including falling back to $MAILDIR or $HOME/mail if a path
        is not specified and using $HOME/<path> if path is relative.
        """
        default_path = os.environ.get('MAILDIR', '{}/mail'.format(os.environ.get('HOME')))
        db_path = notmuch_settings.get('database', 'path', fallback=default_path)

        # If path is relative, notmuch prepends $HOME in front
        if not os.path.isabs(db_path):
            db_path = '{}/{}'.format(os.environ.get('HOME'), db_path)

        return db_path

    def __enter__(self):
        """
        Implements the context manager protocol.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Implements the context manager protocol.
        """
        self.close()

    def open(self, rw=False, retry_for=180, retry_delay=1, create=False):
        if rw:
            if self.handle and self.handle.mode == notmuch.Database.MODE.READ_WRITE:
                return self.handle

            start_time = time.time()
            while True:
                try:
                    self.handle = notmuch.Database(self.db_path,
                                                   mode=notmuch.Database.MODE.READ_WRITE,
                                                   create=create)
                    break
                except notmuch.NotmuchError:
                    time_left = int(retry_for - (time.time() - start_time))

                    if time_left <= 0:
                        raise

                    if time_left % 15 == 0:
                        logging.debug(
                            'Opening the database failed. Will keep trying for another {} seconds'.format(time_left))

                    time.sleep(retry_delay)
        else:
            if not self.handle:
                self.handle = notmuch.Database(self.db_path, create=create)

        return self.handle

    def close(self):
        """
        Closes the notmuch database if it has been opened.
        """
        if self.handle:
            self.handle.close()
            self.handle = None

    def do_query(self, query):
        """
        Executes a notmuch query.

        :param query: the query to execute
        :type  query: str
        :returns: the query result
        :rtype:   :class:`notmuch.Query`
        """
        logging.debug('Executing query %r' % query)
        return notmuch.Query(self.open(), query)

    def get_messages(self, query, full_thread=False):
        """
        Get all messages mathing the given query.

        :param query: the query to execute using :func:`Database.do_query`
        :type  query: str
        :param full_thread: return all messages from mathing threads
        :type  full_thread: bool
        :returns: an iterator over :class:`notmuch.Message` objects
        """
        if not full_thread:
            for message in self.do_query(query).search_messages():
                yield message
        else:
            for thread in self.do_query(query).search_threads():
                for message in self.walk_thread(thread):
                    yield message

    def walk_replies(self, message):
        """
        Returns all replies to the given message.

        :param message: the message to start from
        :type  message: :class:`notmuch.Message`
        :returns: an iterator over :class:`notmuch.Message` objects
        """
        yield message

        # TODO: bindings are *very* unpythonic here... iterator *or* None
        #       is a nono
        replies = message.get_replies()
        if replies != None:
            for message in replies:
                # TODO: yield from
                for message in self.walk_replies(message):
                    yield message

    def walk_thread(self, thread):
        """
        Returns all messages in the given thread.

        :param thread: the tread you are interested in
        :type  thread: :class:`notmuch.Thread`
        :returns: an iterator over :class:`notmuch.Message` objects
        """
        for message in thread.get_toplevel_messages():
            # TODO: yield from
            for message in self.walk_replies(message):
                yield message

    def add_message(self, path, sync_maildir_flags=False, new_mail_handler=None):
        """
        Adds the given message to the notmuch index.

        :param path: path to the message
        :type  path: str
        :param sync_maildir_flags: if `True` notmuch converts the
                                   standard maildir flags to tags
        :type  sync_maildir_flags: bool
        :param new_mail_handler: callback for new messages
        :type  new_mail_handler: a function that is called with a
                                 :class:`notmuch.Message` object as
                                 its only argument
        :raises: :class:`notmuch.NotmuchError` if adding the message fails
        :returns: a :class:`notmuch.Message` object
        """
        # TODO: it would be nice to update notmuchs directory index here
        handle = self.open(rw=True)
        if hasattr(notmuch.Database, 'index_file'):
            message, status = handle.index_file(path, sync_maildir_flags=sync_maildir_flags)
        else:
            message, status = handle.add_message(path, sync_maildir_flags=sync_maildir_flags)

        if status != notmuch.STATUS.DUPLICATE_MESSAGE_ID:
            logging.info('Found new mail in {}'.format(path))

            for tag in get_notmuch_new_tags():
                message.add_tag(tag)

            if new_mail_handler:
                new_mail_handler(message)

        return message

    def remove_message(self, path):
        """
        Remove the given message from the notmuch index.

        :param path: path to the message
        :type  path: str
        """
        self.open(rw=True).remove_message(path)



# SPDX-License-Identifier: ISC
# Copyright (c) 2013 Patrick Gerken <do3cc@patrick-gerken.de>

import unittest


class TestFilterRegistry(unittest.TestCase):

    def test_all_filters_exist(self):
        from afew import FilterRegistry
        self.assertTrue(hasattr(FilterRegistry.all_filters, 'get'))

    def test_entry_point_registration(self):
        from afew import FilterRegistry

        class FakeRegistry:
            name = 'test'

            def load(self):
                return 'class'

        registry = FilterRegistry.FilterRegistry([FakeRegistry()])

        self.assertEqual('class', registry['test'])

    def test_add_FilterRegistry(self):
        from afew import FilterRegistry
        try:
            FilterRegistry.all_filters['test'] = 'class'
            self.assertEqual('class', FilterRegistry.all_filters['test'])
        finally:
            del FilterRegistry.all_filters['test']



# SPDX-License-Identifier: ISC
# Copyright (c) Justus Winter <4winter@informatik.uni-hamburg.de>

import os
import re
import collections
import shlex

from afew.configparser import ConfigParser
from afew.FilterRegistry import all_filters

user_config_dir = os.path.join(os.environ.get('XDG_CONFIG_HOME',
                                              os.path.expanduser('~/.config')),
                               'afew')
user_config_dir = os.path.expandvars(user_config_dir)

settings = ConfigParser()
# preserve the capitalization of the keys.
settings.optionxform = str

settings.readfp(open(os.path.join(os.path.dirname(__file__), 'defaults', 'afew.config')))
settings.read(os.path.join(user_config_dir, 'config'))

# All the values for keys listed here are interpreted as ;-delimited lists
value_is_a_list = ['tags', 'tags_blacklist']
mail_mover_section = 'MailMover'

section_re = re.compile(r'^(?P<name>[a-z_][a-z0-9_]*)(\((?P<parent_class>[a-z_][a-z0-9_]*)\)|\.(?P<index>\d+))?$', re.I)


def get_filter_chain(database):
    filter_chain = []

    for section in settings.sections():
        if section == 'global' or section == mail_mover_section:
            continue

        match = section_re.match(section)
        if not match:
            raise SyntaxError('Malformed section title %r.' % section)

        kwargs = dict(
            (key, settings.get(section, key))
            if key not in value_is_a_list else
            (key, settings.get_list(section, key))
            for key in settings.options(section)
        )

        if match.group('parent_class'):
            try:
                parent_class = all_filters[match.group('parent_class')]
            except KeyError:
                raise NameError(
                    'Parent class %r not found in filter type definition %r.' % (match.group('parent_class'), section))

            new_type = type(match.group('name'), (parent_class,), kwargs)
            all_filters[match.group('name')] = new_type
        else:
            try:
                klass = all_filters[match.group('name')]
            except KeyError:
                raise NameError('Filter type %r not found.' % match.group('name'))
            filter_chain.append(klass(database, **kwargs))

    return filter_chain


def get_mail_move_rules():
    rule_pattern = re.compile(r"'(.+?)':((?P<quote>['\"])(.*?)(?P=quote)|\S+)")
    if settings.has_option(mail_mover_section, 'folders'):
        all_rules = collections.OrderedDict()

        for folder in shlex.split(settings.get(mail_mover_section, 'folders')):
            if settings.has_option(mail_mover_section, folder):
                rules = collections.OrderedDict()
                raw_rules = re.findall(rule_pattern,
                                       settings.get(mail_mover_section, folder))
                for rule in raw_rules:
                    query = rule[0]
                    destination = rule[3] or rule[1]
                    rules[query] = destination
                all_rules[folder] = rules
            else:
                raise NameError("No rules specified for maildir '{}'.".format(folder))

        return all_rules
    else:
        raise NameError("No folders defined to move mails from.")


def get_mail_move_age():
    max_age = 0
    if settings.has_option(mail_mover_section, 'max_age'):
        max_age = settings.get(mail_mover_section, 'max_age')
    return max_age


def get_mail_move_rename():
    rename = False
    if settings.has_option(mail_mover_section, 'rename'):
        rename = settings.get(mail_mover_section, 'rename').lower() == 'true'
    return rename



# SPDX-License-Identifier: ISC
# Copyright (c) Justus Winter <4winter@informatik.uni-hamburg.de>

import os
import glob

__all__ = list(filename[:-3]
               for filename in glob.glob1(os.path.dirname(__file__), '*.py')
               if filename != '__init__.py')



# Copyright (c) Jens Neuhalfen <jens@neuhalfen.name>

import re
from itertools import chain

from afew.filters.BaseFilter import Filter


def _flatten(listOfLists):
    "Flatten one level of nesting"
    return chain.from_iterable(listOfLists)


class PropagateTagsByRegexInThreadFilter(Filter):
    """
    This filter enables a very easy workflow where entire threads can be tagged automatically.

    Assuming the following workflow: all messages for projects or releases should be tagged
    as "project/A", "project/B" respectively "release/1.0.1" or "release/1.2.0".

    In most cases replies to messages retain their context: the project, the release(s), ..

    The following config will propagate all project/... or release/... tags from a thread
    to all new messages.

        [PropagateTagsByRegexInThreadFilter.1]
        propagate_tags = project/.*
        # do not tag spam
        filter = not is:spam

        [PropagateTagsByRegexInThreadFilter.2]
        propagate_tags = release/.*

    Implementation spec: This filter will search through all (new) messages matched by ``filter``.
    For each message ``m`` found it goes through the messages thread an collects all assigned
    tags that match the regexp ``propagate_tags`` (``t``).

    All matching tags ``t`` are then assigned to the new message.
    """

    def handle_message(self, message):
        thread_query = 'thread:"%s"' % (message.get_thread_id(),)
        if self._filter:
            query = self.database.get_messages("(%s) AND (%s)" % (thread_query, self._filter))
        else:
            query = self.database.get_messages(thread_query)

        # the query can only be iterated once, then it is exhausted
        # https://git.notmuchmail.org/git?p=notmuch;a=blob;f=bindings/python/notmuch/messages.py;h=cae5da508f353f12cca585cb056c0b9ed92e29b3;hb=HEAD
        messages = list(query)

        # flatten  tags
        tags_in_thread_t = {m.get_tags() for m in messages}  # a set of Tags instances
        tags_in_thread = set(_flatten(tags_in_thread_t))

        # filter tags
        propagatable_tags_in_thread = {tag for tag in tags_in_thread if self._propagate_tags.fullmatch(tag)}

        if len(propagatable_tags_in_thread):
            self.add_tags(message, *propagatable_tags_in_thread)

    def __init__(self, database, propagate_tags, filter=None, **kwargs):
        if filter:
            self.message = "Propagating tag(s) matching regexp /%s/ from threads to (new) messages matching '%s'" % (
                propagate_tags, filter)
        else:
            self.message = "Propagating tag(s) matching regexp /%s/ from threads to (new) messages'" % (
                propagate_tags,)

        self._filter = filter
        self._propagate_tags = re.compile(propagate_tags)

        super(PropagateTagsByRegexInThreadFilter, self).__init__(database, **kwargs)



# SPDX-License-Identifier: ISC
# Copyright (c) Justus Winter <4winter@informatik.uni-hamburg.de>

import os

from setuptools import setup, find_packages


def get_requires():
    if os.environ.get('TRAVIS') != 'true' and os.environ.get('READTHEDOCS') != 'True':
        yield 'notmuch'
    yield 'chardet'
    yield 'dkimpy'


setup(
    name='afew',
    use_scm_version={'write_to': 'afew/version.py'},
    description="An initial tagging script for notmuch mail",
    url="https://github.com/afewmail/afew",
    license="ISC",
    setup_requires=['setuptools_scm'],
    packages=find_packages(),
    test_suite='afew.tests',
    package_data={
        'afew': ['defaults/afew.config']
    },
    entry_points={
        'console_scripts': [
            'afew = afew.commands:main'],
        'afew.filter': [
            'Filter = afew.filters.BaseFilter:Filter',
            'ArchiveSentMailsFilter = afew.filters.ArchiveSentMailsFilter:ArchiveSentMailsFilter',
            'DKIMValidityFilter = afew.filters.DKIMValidityFilter:DKIMValidityFilter',
            'DMARCReportInspectionFilter = afew.filters.DMARCReportInspectionFilter:DMARCReportInspectionFilter',
            'FolderNameFilter = afew.filters.FolderNameFilter:FolderNameFilter',
            'HeaderMatchingFilter = afew.filters.HeaderMatchingFilter:HeaderMatchingFilter',
            'InboxFilter = afew.filters.InboxFilter:InboxFilter',
            'KillThreadsFilter = afew.filters.KillThreadsFilter:KillThreadsFilter',
            'ListMailsFilter = afew.filters.ListMailsFilter:ListMailsFilter',
            'MeFilter = afew.filters.MeFilter:MeFilter',
            'SentMailsFilter = afew.filters.SentMailsFilter:SentMailsFilter',
            'SpamFilter = afew.filters.SpamFilter:SpamFilter',
            'PropagateTagsByRegexInThreadFilter = afew.filters.PropagateTagsByRegexInThreadFilter:PropagateTagsByRegexInThreadFilter',
            'PropagateTagsInThreadFilter = afew.filters.PropagateTagsInThreadFilter:PropagateTagsInThreadFilter',
        ],
    },
    install_requires=list(get_requires()),
    tests_require=['freezegun'],
    provides=['afew'],
    classifiers=[
        'License :: OSI Approved :: ISC License (ISCL)',
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'Programming Language :: Python',
        'Topic :: Communications :: Email',
        'Topic :: Communications :: Email :: Filters',
        'Topic :: Utilities',
        'Topic :: Database',
    ],
)



# SPDX-License-Identifier: ISC
# Copyright (c) Justus Winter <4winter@informatik.uni-hamburg.de>

import sys

from afew.MailMover import MailMover

try:
    from .files import watch_for_new_files, quick_find_dirs_hack
except ImportError:
    watch_available = False
else:
    watch_available = True


def main(options, database, query_string):
    if options.tag:
        for filter_ in options.enable_filters:
            filter_.run(query_string)
            filter_.commit(options.dry_run)
    elif options.watch:
        if not watch_available:
            sys.exit('Sorry, this feature requires Linux and pyinotify')
        watch_for_new_files(options, database,
                            quick_find_dirs_hack(database.db_path))
    elif options.move_mails:
        for maildir, rules in options.mail_move_rules.items():
            mover = MailMover(options.mail_move_age, options.mail_move_rename, options.dry_run, options.notmuch_args)
            mover.move(maildir, rules)
            mover.close()
    else:
        sys.exit('Weird... please file a bug containing your command line.')



# SPDX-License-Identifier: ISC
# Copyright (c) Justus Winter <4winter@informatik.uni-hamburg.de>

import re

from afew.filters.BaseFilter import Filter
from afew.NotmuchSettings import notmuch_settings


class SentMailsFilter(Filter):
    message = 'Tagging all mails sent by myself to others'
    _bare_email_re = re.compile(r"[^<]*<(?P<email>[^@<>]+@[^@<>]+)>")

    def __init__(self, database, sent_tag='', to_transforms=''):
        super().__init__(database)

        my_addresses = set()
        my_addresses.add(notmuch_settings.get('user', 'primary_email'))
        if notmuch_settings.has_option('user', 'other_email'):
            for other_email in notmuch_settings.get_list('user', 'other_email'):
                my_addresses.add(other_email)

        self.query = (
                '(' +
                ' OR '.join('from:"%s"' % address for address in my_addresses)
                + ') AND NOT (' +
                ' OR '.join('to:"%s"' % address for address in my_addresses)
                + ')'
        )

        self.sent_tag = sent_tag
        self.to_transforms = to_transforms
        if to_transforms:
            self.__email_to_tags = self.__build_email_to_tags(to_transforms)

    def handle_message(self, message):
        if self.sent_tag:
            self.add_tags(message, self.sent_tag)
        if self.to_transforms:
            for header in ('To', 'Cc', 'Bcc'):
                email = self.__get_bare_email(message.get_header(header))
                for tag in self.__pick_tags(email):
                    self.add_tags(message, tag)
                else:
                    break

    def __build_email_to_tags(self, to_transforms):
        email_to_tags = dict()

        for rule in to_transforms.split():
            if ':' in rule:
                email, tags = rule.split(':')
                email_to_tags[email] = tuple(tags.split(';'))
            else:
                email = rule
                email_to_tags[email] = tuple()

        return email_to_tags

    def __get_bare_email(self, email):
        if not '<' in email:
            return email
        else:
            match = self._bare_email_re.search(email)
            return match.group('email')

    def __pick_tags(self, email):
        if email in self.__email_to_tags:
            tags = self.__email_to_tags[email]
            if tags:
                return tags
            else:
                user_part, domain_part = email.split('@')
                return (user_part,)

        return tuple()



# SPDX-License-Identifier: ISC
# Copyright (c) Amadeusz Zolnowski <aidecoe@aidecoe.name>

"""
DMARC report inspection filter.

Looks into DMARC report whether all results are successful or any is failing.
Add tags 2 of the tags below:
- dmarc/dkim-ok
- dmarc/dkim-fail
- dmarc/spf-ok
- dmarc/spf-fail

"""

import logging
import re
import tempfile
import xml.etree.ElementTree as ET
import zipfile

from .BaseFilter import Filter


class DMARCInspectionError(Exception):
    """Failed to inspect DMARC report.
    """


class ReportFilesIterator:
    """
    Iterator over DMARC reports files attached to the e-mail either directly or
    in ZIP files.

    Returns content of each document file (as bytes, not as string) which needs
    to be decoded from charset encoding.
    """
    def __init__(self, message):
        self.message = message

    def __iter__(self):
        for part in self.message.get_message_parts():
            if part.get_content_type() == 'application/zip':
                with tempfile.TemporaryFile(suffix='.zip') as file:
                    file.write(part.get_payload(decode=True))
                    try:
                        with zipfile.ZipFile(file) as zip_file:
                            for member_file in zip_file.infolist():
                                if member_file.filename.endswith('.xml'):
                                    yield zip_file.read(member_file)
                    except zipfile.BadZipFile as zip_error:
                        raise DMARCInspectionError(str(zip_error)) \
                            from zip_error
            elif part.get_content_type() == 'application/xml':
                yield part.get_payload(decode=True)


def and_dict(dict1, dict2):
    """
    Apply logical conjunction between values of dictionaries of the same keys.

    Keys set must be identical in both dictionaries. Otherwise KeyError
    exception is raised.

    :param dict1: Dictionary of bool values.
    :param dict2: Dictionary of bool values.
    :returns: A dictionary with the same set of keys but with modified values.
    """
    dict3 = {}
    for key in dict1.keys():
        dict3[key] = dict1[key] & dict2.get(key, False)
    return dict3


def has_failed(node):
    """
    Check whether status is "failed".

    To avoid false positives check whether status is one of "pass" or "none".

    :param node: XML node holding status as text.
    :returns: Whether the status is reported as "failed".
    """
    return (node.text.strip() not in ['pass', 'none'])


def read_auth_results(document):
    """
    Parse DMARC document.

    Look for results for DKIM and SPF. If there's more than one record, return
    `True` only and only if all of the records of particular type (DKIM or SPF)
    are "pass".

    :returns: Results as a dictionary where keys are: `dkim` and `spf` and
    values are boolean values.
    """
    try:
        results = {'dkim': True, 'spf': True}
        root = ET.fromstring(document)
        for record in root.findall('record'):
            auth_results = record.find('auth_results')
            if auth_results:
                dkim = auth_results.find('dkim')
                if dkim:
                    dkim = dkim.find('result')
                    results['dkim'] &= not has_failed(dkim)
                spf = auth_results.find('spf')
                if spf:
                    spf = spf.find('result')
                    results['spf'] &= not has_failed(spf)
    except ET.ParseError as parse_error:
        raise DMARCInspectionError(str(parse_error)) from parse_error

    return results


class DMARCReportInspectionFilter(Filter):
    """
    Inspect DMARC reports for DKIM and SPF status.
    """
    def __init__(self,                     # pylint: disable=too-many-arguments
                 database,
                 dkim_ok_tag='dmarc/dkim-ok',
                 dkim_fail_tag='dmarc/dkim-fail',
                 spf_ok_tag='dmarc/spf-ok',
                 spf_fail_tag='dmarc/spf-fail'):
        super().__init__(database)
        self.dkim_tag = {True: dkim_ok_tag, False: dkim_fail_tag}
        self.spf_tag = {True: spf_ok_tag, False: spf_fail_tag}
        self.dmarc_subject = re.compile(r'^report domain:',
                                        flags=re.IGNORECASE)
        self.log = logging.getLogger('{}.{}'.format(
            self.__module__, self.__class__.__name__))

    def handle_message(self, message):
        if not self.dmarc_subject.match(message.get_header('Subject')):
            return

        auth_results = {'dkim': True, 'spf': True}

        try:
            for file_content in ReportFilesIterator(message):
                document = file_content.decode('UTF-8')
                auth_results = and_dict(auth_results,
                                        read_auth_results(document))

            self.add_tags(message,
                          'dmarc',
                          self.dkim_tag[auth_results['dkim']],
                          self.spf_tag[auth_results['spf']])
        except DMARCInspectionError as inspection_error:
            self.log.error(
                "Failed to verify DMARC report of '%s': %s (not tagging)",
                message.get_message_id(),
                inspection_error
            )
